{% extends "base.html" %}

{% block announce %}
Version 0.1.0-pre.3. This library is still under active development. Expect bugs and breaking changes.
{% endblock %}

<!-- Social cars -->
{% block extrahead %}
    {% set title = config.site_name %}
    {% if page and page.meta and page.meta.title %}
        {% set title = title ~ " - " ~ page.meta.title %}
    {% elif page and page.title and not page.is_homepage %}
        {% set title = title ~ " - " ~ page.title | striptags %}
    {% endif %}
<meta property="og:type" content="website" />
<meta property="og:title" content="{{ title }}" />
<meta property="og:description" content="{{ config.site_description }}" />
<meta property="og:url" content="{{ page.canonical_url }}" />
<meta property="og:image" content="https://ioverho.github.io/prob_conf_mat/assets/social-preview.png" />
<meta property="og:image:type" content="image/png" />
<meta property="og:image:width" content="1200" />
<meta property="og:image:height" content="630" />
{% endblock %}
